import uuid

# Battery Service
batteryServiceUUID = uuid.UUID("0000180F-0000-1000-8000-00805f9b34fb")
batteryLevelCharacteristicUUID = uuid.UUID("00002A19-0000-1000-8000-00805f9b34fb")

# Device Information Service
deviceInformationServiceUUID = uuid.UUID("0000180a-0000-1000-8000-00805f9b34fb")
manufacturerNameUUID = uuid.UUID("00002A29-0000-1000-8000-00805F9B34FB")
modelNumberUUID = uuid.UUID("00002A24-0000-1000-8000-00805F9B34FB")
serialNumberUUID = uuid.UUID("00002A25-0000-1000-8000-00805F9B34FB")
firmwareVersionUUID = uuid.UUID("00002A26-0000-1000-8000-00805F9B34FB")
hardwareVersionUUID = uuid.UUID("00002A27-0000-1000-8000-00805F9B34FB")
softwareVersionUUID = uuid.UUID("00002A28-0000-1000-8000-00805F9B34FB")

# Apogee Service
apogeeServiceUUID = uuid.UUID("B3E00001-2594-42A1-A5FE-4E660FF2868F")
liveDataUUID = uuid.UUID("B3E00002-2594-42A1-A5FE-4E660FF2868F")
sensorIDUUID = uuid.UUID("B3E00003-2594-42A1-A5FE-4E660FF2868F")
aliasUUID = uuid.UUID("B3E00004-2594-42A1-A5FE-4E660FF2868F")
liveDataAverageUUID = uuid.UUID("B3E00005-2594-42A1-A5FE-4E660FF2868F")
LEDStateUUID = uuid.UUID("B3E00006-2594-42A1-A5FE-4E660FF2868F")
fanControlUUID = uuid.UUID("B3E00007-2594-42A1-A5FE-4E660FF2868F")
dataLoggingControlUUID = uuid.UUID("B3E00010-2594-42A1-A5FE-4E660FF2868F")
dataLoggingIntervalsUUID = uuid.UUID("B3E00012-2594-42A1-A5FE-4E660FF2868F")
dataLogTransferUUID = uuid.UUID("B3E00013-2594-42A1-A5FE-4E660FF2868F")
dataLogCollectionRateUUID = uuid.UUID("B3E00014-2594-42A1-A5FE-4E660FF2868F")
coefficient1UUID = uuid.UUID("B3E00100-2594-42A1-A5FE-4E660FF2868F")
coefficient2UUID = uuid.UUID("B3E00101-2594-42A1-A5FE-4E660FF2868F")
modbusSettingsUUID = uuid.UUID("B3E00200-2594-42A1-A5FE-4E660FF2868F")
currentTimeUUID = uuid.UUID("B3E0000A-2594-42A1-A5FE-4E660FF2868F")
timeTillFullUUID = uuid.UUID("B3E0000C-2594-42A1-A5FE-4E660FF2868F")
logEntriesUUID = uuid.UUID("B3E0000D-2594-42A1-A5FE-4E660FF2868F")
lastTransferredTimestampUUID = uuid.UUID("B3E0000E-2594-42A1-A5FE-4E660FF2868F")
darkOffsetUUID = uuid.UUID("B3E000FF-2594-42A1-A5FE-4E660FF2868F")